/*
 *	Frame buffer driver for the Cyrix MediaGX/MediaGXm
 *	machines with 5510,5520,5530video support logic. The
 *	5520 is the only one currently tested and not all the
 *	5510 DAC programming logic is included right now.
 *
 *		Alan Cox <alan@redhat.com>
 *
 * Based on:
 * framebuffer driver for VBE 2.0 compliant graphic boards
 * (c) 1998 Gerd Knorr <kraxel@cs.tu-berlin.de>
 * 
 * and
 * Emmanuel Marty's <core@suntech.fr> KGI driver for this 
 * device.
 */ 

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/malloc.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/console.h>
#include <linux/selection.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/pci.h>

#include <asm/io.h>

#include "fbcon.h"
#include "fbcon-cfb8.h"
#include "fbcon-cfb16.h"

#include "cx5520.h"
#include "gx.h"

/*
 * card parameters
 */

/* card */

char *video_base;
int   video_size;
char *video_vbase;        /* mapped */
char *vcomp_base;	/* Compression buffers */
char *vcomp_vbase;	

/* mode */
int  video_bpp = 8;
int  video_width = 1024;
int  video_height = 768;
int  video_type = FB_TYPE_PACKED_PIXELS;
int  video_visual;
int  video_linelength;
int  video_cmap_len;

/* --------------------------------------------------------------------- */

static struct fb_var_screeninfo mediagxfb_defined = {
	0,0,0,0,	/* W,H, W, H (virtual) load xres,xres_virtual*/
	0,0,		/* virtual -> visible no offset */
	8,		/* depth -> load bits_per_pixel */
	0,		/* greyscale ? */
	{0,0,0},	/* R */
	{0,0,0},	/* G */
	{0,0,0},	/* B */
	{0,0,0},	/* transparency */
	0,		/* standard pixel format */
	FB_ACTIVATE_NOW,
	-1,-1,
	0,
	0L,0L,0L,0L,0L,
	0L,0L,0,	/* No sync info */
	FB_VMODE_NONINTERLACED,
	{0,0,0,0,0,0}
};

static struct display disp;
static struct fb_info fb_info;
static struct { u_char red, green, blue, pad; } palette[256];

static int inverse = 0;

static int currcon = 0;

/* --------------------------------------------------------------------- */
/* speed up scrolling                                                    */

#define USE_REDRAW   1
#define USE_MEMMOVE  2
  
static int mediagxfb_scroll = USE_REDRAW;

static void *m5520_base;
static u32 m5520_pbase;
static void *mediagx_base;
static int support_chip = 0;

struct cyrix_mode_data
{
        int xres;
        int yres;
        int bpp;
        int hz;

        /* MediaGX controller values */

        unsigned long gcfg;
        unsigned long tcfg;
        unsigned long ocfg;
        unsigned long fb_offset;
        unsigned long cb_offset;
        unsigned long curs_offset;
        unsigned long line_delta;
        unsigned long buffer_size;
        unsigned long htiming1;
        unsigned long htiming2;
        unsigned long htiming3;
        unsigned long fp_htiming;
        unsigned long vtiming1;
        unsigned long vtiming2;
        unsigned long vtiming3;
        unsigned long fp_vtiming;

	/* RAMDAC controller values */

        unsigned char ics5342_m;
        unsigned char ics5342_n;

	/* CX5520 clock value */
	
        unsigned long cx5520_clock;

};

/*
 *	These values are defined by the Cyrix example programs
 */
 
static struct cyrix_mode_data modes[4] = 
{
	{
		640, 480, 8, 60,                                      // display parameters
		0x00006541, 0x0000036F, 0x00003005,                   // gcfg, tcfg, ocfg
		0x00000000, 0x00000280, 0x001FFF00,                   // memory organization
		0x00100100, 0x00008250,                               // line delta, buffer size
		/* 799, 639   791, 647   743, 655     743, 655 */
		0x031F027F, 0x03170287, 0x02E7028F, 0x02E7028F,       // horizontal timings
		/* 524, 479   516,487    491, 498     491,489 */
		0x020C01DF, 0x020401E7, 0x01EB01E9, 0x01EB01E9,       // vertical timings
		0x7D, 0x47,                                           // ICS5342 = 50.350 MHz
		0x37058AFF,                                           // CX5520 = 50.350 MHz
	},

	{
		800, 600, 8, 60,                                      // display parameters
		0x00006541, 0x0000006F, 0x00003005,                   // gcfg, tcfg, ocfg
		0x00000000, 0x00000320, 0x001FFF00,                   // memory organization
		0x00100100, 0x00006A64,                               // line delta, buffer size
		/* 1055, 799  1047, 807   975, 839    975, 839 */
		0x041F031F, 0x04170327, 0x03CF0347, 0x03CF0347,       // horizontal timings
		/* 627, 599   627, 599    604, 600    604, 600 */
		0x02730257, 0x02730257, 0x025C0258, 0x025C0258,       // vertical timings
		0x79, 0x29,                                           // ICS5342 = 80.000 MHz
		0x271B1DFF,                                           // CX5520 = 80.000 MHz
	},

	{
		1024, 768, 8, 60,                                     // display parameters
		0x00006581, 0x0000006F, 0x00003005,                   // gcfg, tcfg, ocfg
		0x00000000, 0x000C0000, 0x001FFF00,                   // memory organization
		0x00044100, 0x00008280,                               // line delta, buffer size
		/* 1343, 1023 1343, 1023  1183, 1047  1183, 1047 */
		0x053F03FF, 0x053F03FF, 0x049F0417, 0x049F0417,       // horizontal timings
		/* 805, 767  805, 767     776, 770    776, 770 */
		0x032502FF, 0x032502FF, 0x03080302, 0x03080302,       // vertical timings
		0x77, 0x29,                                           // ICS5342 = 78.750 MHz
		0x27915BFF,                                           // CX5520 = 78.750 MHz
	},
	/* Slot 3 we fill with our own saved data */

};

#define NUM_DISPLAY_MODES 3

extern inline void mediagx_write_reg32(u8 *a, int b, u32 v)
{
	volatile u32 *vp=(volatile u32 *)(a+b);
	*(u32 *)vp=v;
}

extern inline u32 mediagx_read_reg32(u8 *a, int b)
{
	volatile u32 *vp=(volatile u32 *)(a+b);
	return *((u32 *)vp);
}

/*
 *	The original MediaGX has an external ramdac so we have to 
 *	go through the unsual contortions in programming these horrors
 */

void ramdac_write(unsigned long value)
{
        mediagx_write_reg32(mediagx_base, MC_RAMDAC_ACC, value);
}

/*
 *	Set the index into the ramdac
 */
 
static void ramdac_select(unsigned long index)
{
        u32 unlock, value;

        unlock = mediagx_read_reg32(mediagx_base, DC_UNLOCK);
        mediagx_write_reg32(mediagx_base, DC_UNLOCK, DC_UNLOCK_VALUE);

        mediagx_read_reg32(mediagx_base, MC_RAMDAC_ACC);

	/* Register select links */
        value = mediagx_read_reg32(mediagx_base, DC_GENERAL_CFG);
	value = (value & ~DC_GCFG_DAC_RS_MASK) | (index << DC_GCFG_DAC_RS_POS);
        mediagx_write_reg32(mediagx_base, DC_GENERAL_CFG, value);

        mediagx_write_reg32(mediagx_base, DC_UNLOCK, unlock);
}

/*
 *	Propogate the palette onto the RAMDAC
 */

static void ramdac_palette(u8 index, u8 r, u8 g, u8 b)
{
        unsigned long unlock;

        unlock = mediagx_read_reg32(mediagx_base, DC_UNLOCK);
        mediagx_write_reg32(mediagx_base, DC_UNLOCK, DC_UNLOCK_VALUE);

	/* Colour number */
	
        ramdac_select(0);
        ramdac_write(index);

	/* Begin write */
        ramdac_select(1);

 	ramdac_write(r<<3);
 	ramdac_write(g<<3);
 	ramdac_write(b<<3);

        mediagx_write_reg32(mediagx_base, DC_UNLOCK, unlock);
}

/*
 *	Load the clocks onto the RAMDAC
 */
 
void program_ics5342_ramdac(u32 width, u32 depth, u32 ics5342_m, u32 ics5342_n)
{
        unsigned long unlock;

        unlock = mediagx_read_reg32(mediagx_base, DC_UNLOCK);
        mediagx_write_reg32(mediagx_base, DC_UNLOCK, DC_UNLOCK_VALUE);

	/*
	 *	Pixel mask register
	 */
	 
        ramdac_select(2);
        ramdac_write(0xFF);
        
	/* Index 0 first */
	
        ramdac_select(4);
        ramdac_write(0x0E);

        ramdac_select(5);
        ramdac_write(0x20);

	/* PLL Index 2 */        

        ramdac_select(4);
        ramdac_write(0x02);
        ramdac_select(5);
        ramdac_write(ics5342_m);
        ramdac_write(ics5342_n);

	/* PLL write address 14 */
	
        ramdac_select(4);
        ramdac_write(0x0E);

	/* PLL index 2 */
        
        ramdac_select(5);
        ramdac_write(0x22);

	/* DAC mode */
        
        ramdac_select(6);

	if (depth == 16)
	{
        	if (width < 1024)
			ramdac_write(0xC0);
                else
			ramdac_write(0x50);
	}
	else
	{
        	ramdac_write(0x00);
	}
	ramdac_select(0);
	mediagx_write_reg32(mediagx_base, DC_UNLOCK, unlock);
}

/*
 *	Set the mode
 */
 
void mediagx_video_mode(int size)
{
	unsigned long unlock;
        unsigned long gcfg, tcfg, ocfg, dcfg;
        struct cyrix_mode_data *pMode = &modes[size];

        unlock = mediagx_read_reg32(mediagx_base, DC_UNLOCK);
        mediagx_write_reg32(mediagx_base, DC_UNLOCK, DC_UNLOCK_VALUE);

	gcfg = mediagx_read_reg32(mediagx_base, DC_GENERAL_CFG);
	tcfg = mediagx_read_reg32(mediagx_base, DC_TIMING_CFG);
	ocfg = mediagx_read_reg32(mediagx_base, DC_OUTPUT_CFG);

	if (support_chip == 2)
	{
		dcfg = mediagx_read_reg32(m5520_base, CX5520_DISPLAY_CONFIG);
                dcfg &= ~(unsigned long) CX5520_DCFG_DAC_BL_EN;
		mediagx_write_reg32(m5520_base, CX5520_DISPLAY_CONFIG, dcfg);
        }

	/* Image off */

	tcfg &= ~(unsigned long)DC_TCFG_BLKE;
	mediagx_write_reg32(mediagx_base, DC_TIMING_CFG, tcfg);

 	/* Time generator off */

	tcfg &= ~(unsigned long)DC_TCFG_TGEN;
        mediagx_write_reg32(mediagx_base, DC_TIMING_CFG, tcfg);

	/* Drain the memory FIFO */
 	udelay(2000);

	/* FIFO load off, compression off */

	gcfg &= ~(unsigned long)(DC_GCFG_DFLE | DC_GCFG_CMPE | DC_GCFG_DECE);
	mediagx_write_reg32(mediagx_base, DC_GENERAL_CFG, gcfg);

        /* Dclock multipler clear */

        gcfg &= ~(unsigned long)(DC_GCFG_DDCK | DC_GCFG_DPCK | DC_GCFG_DFCK);
        gcfg &= ~(unsigned long)DC_GCFG_DCLK_MASK;
        mediagx_write_reg32(mediagx_base, DC_GENERAL_CFG, gcfg);

	/*
	 *	Program the dotclock generator
	 */

	if (support_chip==1)
		program_ics5342_ramdac(pMode->xres, pMode->bpp, pMode->ics5342_m, pMode->ics5342_n);

        if (support_chip==2)
	{	
		mediagx_write_reg32(m5520_base, 0x0E14, pMode->cx5520_clock | 0x40000000);
		mediagx_write_reg32(m5520_base, 0x0E14, pMode->cx5520_clock);		
	}
	
	/*
	 *	The dotclock is now unstable for a bit
	 */
	 
	udelay(2000);

	/*
	 *	Set the dot clock multiplier. x1 means x.5 x2 means x1 etc
	 */

        gcfg |= pMode->gcfg & DC_GCFG_DCLK_MASK;
        mediagx_write_reg32(mediagx_base, DC_GENERAL_CFG, gcfg);
        
	/*
	 *	Settle the dot clock
	 */

        udelay(2000);

	/*
	 *	Display controller parameters - timing, sync etc
	 */
	 
        mediagx_write_reg32(mediagx_base, DC_FB_ST_OFFSET, pMode->fb_offset);
        mediagx_write_reg32(mediagx_base, DC_CB_ST_OFFSET, pMode->cb_offset);
        mediagx_write_reg32(mediagx_base, DC_CURS_ST_OFFSET, pMode->curs_offset);
        mediagx_write_reg32(mediagx_base, DC_LINE_DELTA, pMode->line_delta);
        mediagx_write_reg32(mediagx_base, DC_BUF_SIZE, pMode->buffer_size);
        mediagx_write_reg32(mediagx_base, DC_H_TIMING_1, pMode->htiming1);
        mediagx_write_reg32(mediagx_base, DC_H_TIMING_2, pMode->htiming2);
        mediagx_write_reg32(mediagx_base, DC_H_TIMING_3, pMode->htiming3);
        mediagx_write_reg32(mediagx_base, DC_FP_H_TIMING, pMode->fp_htiming);
        mediagx_write_reg32(mediagx_base, DC_V_TIMING_1, pMode->vtiming1);
        mediagx_write_reg32(mediagx_base, DC_V_TIMING_2, pMode->vtiming2);
        mediagx_write_reg32(mediagx_base, DC_V_TIMING_3, pMode->vtiming3);
        mediagx_write_reg32(mediagx_base, DC_FP_V_TIMING, pMode->fp_vtiming);

	/*
	 *	Precomputed data
	 */
	 
        gcfg = pMode->gcfg;
        tcfg = pMode->tcfg;
        ocfg = pMode->ocfg;

        if (support_chip == 2)
        {
        	/*
        	 *	The CX5520 companion does the sync polarities
        	 */

                dcfg &= ~(unsigned long) CX5520_DCFG_CRT_HSYNC_POL;
                dcfg &= ~(unsigned long) CX5520_DCFG_CRT_VSYNC_POL;
                dcfg &= ~(unsigned long) CX5520_DCFG_FP_HSYNC_POL;
                dcfg &= ~(unsigned long) CX5520_DCFG_FP_VSYNC_POL;
                if (tcfg & DC_TCFG_CHSP)
                	dcfg |= CX5520_DCFG_CRT_HSYNC_POL;
                if (tcfg & DC_TCFG_CVSP)
                	dcfg |= CX5520_DCFG_CRT_VSYNC_POL;
                if (tcfg & DC_TCFG_FHSP)
                	dcfg |= CX5520_DCFG_FP_HSYNC_POL;
                if (tcfg & DC_TCFG_FVSP)
                	dcfg |= CX5520_DCFG_FP_VSYNC_POL;
                tcfg &= ~(unsigned long) (DC_TCFG_CHSP | DC_TCFG_CVSP);
                tcfg &= ~(unsigned long) (DC_TCFG_FHSP | DC_TCFG_FVSP);
        }

	/*
	 *	Load configuration registers. The limited info from Cyrix
	 *	says we must use this order
	 */
	 
        mediagx_write_reg32(mediagx_base, DC_OUTPUT_CFG, pMode->ocfg);
        mediagx_write_reg32(mediagx_base, DC_TIMING_CFG, pMode->tcfg);
        mediagx_write_reg32(mediagx_base, DC_GENERAL_CFG, pMode->gcfg);
        
        if (support_chip == 2)
        {
		/*
		 *	Turn the display back on
		 */

                dcfg |= CX5520_DCFG_DIS_EN;
                dcfg |= CX5520_DCFG_HSYNC_EN;
                dcfg |= CX5520_DCFG_VSYNC_EN;
                
                /*
                 *	Turn the monitor on
                 */

                dcfg |= CX5520_DCFG_DAC_BL_EN;
                dcfg |= CX5520_DCFG_DAC_PWDNX;

		/*
		 *	LCD panel optionally
		 */
		 
                if (0 /*FlatPanel*/)
                {
                        dcfg |= CX5520_DCFG_FP_PWR_EN;
                        dcfg |= CX5520_DCFG_FP_DATA_EN;
                }

		/*
		 *	We have to specify which stream goes to the palette
		 *	This raises the fascinating unanswered question -
		 *	what are the other streams ??
		 */
		 
                dcfg |= CX5520_DCFG_GV_PAL_BYP;
                mediagx_write_reg32(m5520_base, CX5520_DISPLAY_CONFIG, dcfg);
        }

	/*
	 *	Relock the chip.
	 */
	 
        mediagx_write_reg32(mediagx_base, DC_UNLOCK, unlock);
        return;
}


/* --------------------------------------------------------------------- */

/*
 * Open/Release the frame buffer device
 */

static int mediagxfb_open(struct fb_info *info, int user)
{
	/*
	 * Nothing, only a usage count for the moment
	 */
	MOD_INC_USE_COUNT;
	return(0);
}

static int mediagxfb_release(struct fb_info *info, int user)
{
	MOD_DEC_USE_COUNT;
	return(0);
}

static int fb_update_var(int con, struct fb_info *info)
{
	return 0;
}

static int mediagxfb_get_fix(struct fb_fix_screeninfo *fix, int con,
			 struct fb_info *info)
{
	memset(fix, 0, sizeof(struct fb_fix_screeninfo));
	strcpy(fix->id,"Cyrix MediaGX");

	fix->smem_start=(char *)video_vbase;
	fix->smem_len=video_size;
	fix->type = video_type;
	fix->visual = video_visual;
	fix->xpanstep=0;
	fix->ypanstep=0;
	fix->ywrapstep=0;
	fix->line_length=video_linelength;
	return 0;
}

static int mediagxfb_get_var(struct fb_var_screeninfo *var, int con,
			 struct fb_info *info)
{
	if(con==-1)
		memcpy(var, &mediagxfb_defined, sizeof(struct fb_var_screeninfo));
	else
		*var=fb_display[con].var;
	return 0;
}


static void mediagx_show_pointer(u32 x, u32 y) 
{
	u32 lockval, gcfg;
	
	/* Unlock the display controller registers */
	lockval = mediagx_read_reg32 (mediagx_base, DC_UNLOCK);
	mediagx_write_reg32 (mediagx_base, DC_UNLOCK, DC_UNLOCK_VALUE);

	/* Set cursor position */
	mediagx_write_reg32 (mediagx_base, DC_CURSOR_X, x & 0x7FF);
	mediagx_write_reg32 (mediagx_base, DC_CURSOR_Y, y & 0x3FF);

	/* Enable hardware pointer */
	gcfg = mediagx_read_reg32 (mediagx_base, DC_GENERAL_CFG);
	gcfg = gcfg | DC_GCFG_CURE;
	mediagx_write_reg32 (mediagx_base, DC_GENERAL_CFG, gcfg);

	/* Restore status of lock register */
	mediagx_write_reg32 (mediagx_base, DC_UNLOCK, lockval);
}

static void mediagx_hide_pointer(void) 
{
	u32	lockval, gcfg;

	/* Unlock the display controller registers */
	lockval = mediagx_read_reg32(mediagx_base, DC_UNLOCK);
	mediagx_write_reg32 (mediagx_base, DC_UNLOCK, DC_UNLOCK_VALUE);

	/* Disable hardware pointer */
	gcfg = mediagx_read_reg32 (mediagx_base, DC_GENERAL_CFG);
	gcfg = gcfg & ~DC_GCFG_CURE;
	mediagx_write_reg32 (mediagx_base, DC_GENERAL_CFG, gcfg);

	/* Restore status of lock register */
	mediagx_write_reg32 (mediagx_base, DC_UNLOCK, lockval);
}

static void mediagx_cursor(struct display *p, int mode, int x, int y)
{
	switch(mode)
	{
		case CM_ERASE:
			mediagx_hide_pointer();
			break;
		case CM_MOVE:
		case CM_DRAW:
			mediagx_show_pointer(x,y);
	}	
}


static void mediagxfb_set_disp(int con)
{
	struct fb_fix_screeninfo fix;
	struct display *display;
	struct display_switch *sw=NULL;
	
	if (con >= 0)
		display = &fb_display[con];
	else
		display = &disp;	/* used during initialization */

	mediagxfb_get_fix(&fix, con, 0);

	memset(display, 0, sizeof(struct display));
	display->screen_base = (void *)video_vbase;
	display->visual = fix.visual;
	display->type = fix.type;
	display->type_aux = fix.type_aux;
	display->ypanstep = fix.ypanstep;
	display->ywrapstep = fix.ywrapstep;
	display->line_length = fix.line_length;
	display->next_line = fix.line_length;
	display->can_soft_blank = 0;
	display->inverse = inverse;
	mediagxfb_get_var(&display->var, -1, &fb_info);

	switch (video_bpp) {
#ifdef FBCON_HAS_CFB8
	case 8:
		sw = &fbcon_cfb8;
		printk("Selected 8bit fb ops\n");
		break;
#endif
#ifdef FBCON_HAS_CFB16
	case 15:
	case 16:
		sw = &fbcon_cfb16;
		break;
#endif
	default:
		return;
	}
	display->dispsw = sw;
#if 0
	mediagxfb_sw.cursor = mediagx_cursor;
#endif	
//	if (mediagxfb_scroll == USE_REDRAW) {
//		display->scrollmode = SCROLL_YREDRAW;
//		mediagxfb_sw.bmove = fbcon_redraw_bmove;
//	}
	printk("Going to use mode %d\n", video_width);
	switch(video_width)	/* FIXME */
	{
		case 640:
			mediagx_video_mode(0);
			break;
		case 800:
			mediagx_video_mode(1);
			break;
		case 1024:
			mediagx_video_mode(2);
			break;
	}
	printk("Mode set\n");
}

static int mediagxfb_set_var(struct fb_var_screeninfo *var, int con,
			 struct fb_info *info)
{
	memcpy(var, &mediagxfb_defined, sizeof(struct fb_var_screeninfo));
	return 0;
}

static int mediagx_getcolreg(unsigned regno, unsigned *red, unsigned *green,
			  unsigned *blue, unsigned *transp,
			  struct fb_info *fb_info)
{
	/*
	 *  Read a single color register and split it into colors/transparent.
	 *  Return != 0 for invalid regno.
	 */

	if (regno >= video_cmap_len)
		return 1;

	*red   = palette[regno].red;
	*green = palette[regno].green;
	*blue  = palette[regno].blue;
	return 0;
}

static int mediagx_setcolreg(unsigned regno, unsigned red, unsigned green,
			  unsigned blue, unsigned transp,
			  struct fb_info *fb_info)
{
	/*
	 *  Set a single color register. The values supplied are
	 *  already rounded down to the hardware's capabilities
	 *  (according to the entries in the `var' structure). Return
	 *  != 0 for invalid regno.
	 */
	
	if (regno >= video_cmap_len)
		return 1;
	
	palette[regno].red   = red;
	palette[regno].green = green;
	palette[regno].blue  = blue;
	
	switch (video_bpp) {
#ifdef FBCON_HAS_CFB8
	case 8:
	{
		mediagx_write_reg32(mediagx_base, DC_PAL_ADDRESS, regno);
		mediagx_write_reg32(mediagx_base, DC_PAL_DATA, red<<12 | green <<6 | blue);

		if(support_chip==1)
			ramdac_palette(regno, red, green, blue);
		break;
	}
#endif
#ifdef FBCON_HAS_CFB16
	case 15:
	case 16:
		fbcon_cfb16_cmap[regno] =
			(red << mediagxfb_defined.red.offset) | (green << 5) | blue;
		break;
#endif
    }
    return 0;
}

static void do_install_cmap(int con, struct fb_info *info)
{
	if (con != currcon)
		return;
	if (fb_display[con].cmap.len)
		fb_set_cmap(&fb_display[con].cmap, &fb_display[con].var, 1,
			    mediagx_setcolreg, info);
	else
		fb_set_cmap(fb_default_cmap(video_cmap_len),
			    &fb_display[con].var, 1, mediagx_setcolreg,
			    info);
}

static int mediagxfb_get_cmap(struct fb_cmap *cmap, int kspc, int con,
			   struct fb_info *info)
{
	if (con == currcon) /* current console? */
		return fb_get_cmap(cmap, &fb_display[con].var, kspc, mediagx_getcolreg, info);
	else if (fb_display[con].cmap.len) /* non default colormap? */
		fb_copy_cmap(&fb_display[con].cmap, cmap, kspc ? 0 : 2);
	else
		fb_copy_cmap(fb_default_cmap(video_cmap_len),
		     cmap, kspc ? 0 : 2);
	return 0;
}

static int mediagxfb_set_cmap(struct fb_cmap *cmap, int kspc, int con,
			   struct fb_info *info)
{
	int err;

	if (!fb_display[con].cmap.len) {	/* no colormap allocated? */
		err = fb_alloc_cmap(&fb_display[con].cmap,video_cmap_len,0);
		if (err)
			return err;
	}
	if (con == currcon)			/* current console? */
		return fb_set_cmap(cmap, &fb_display[con].var, kspc, mediagx_setcolreg, info);
	else
		fb_copy_cmap(cmap, &fb_display[con].cmap, kspc ? 0 : 1);
	return 0;
}

static int mediagxfb_pan_display(struct fb_var_screeninfo *var, int con,
			     struct fb_info *info)
{
	/* no panning */
	return -EINVAL;
}

static int mediagxfb_ioctl(struct inode *inode, struct file *file, 
		       unsigned int cmd, unsigned long arg, int con,
		       struct fb_info *info)
{
	return -EINVAL;
}

static struct fb_ops mediagxfb_ops = {
	mediagxfb_open,
	mediagxfb_release,
	mediagxfb_get_fix,
	mediagxfb_get_var,
	mediagxfb_set_var,
	mediagxfb_get_cmap,
	mediagxfb_set_cmap,
	mediagxfb_pan_display,
	mediagxfb_ioctl
};

void mediagxfb_setup(char *options, int *ints)
{
	char *this_opt;
	
	fb_info.fontname[0] = '\0';
	
	if (!options || !*options)
		return;
	
	for(this_opt=strtok(options,","); this_opt; this_opt=strtok(NULL,",")) {
		if (!*this_opt) continue;
		
		printk("mediagxfb_setup: option %s\n", this_opt);
		
		if (! strcmp(this_opt, "inverse"))
			inverse=1;
		else if (! strcmp(this_opt, "redraw"))
			mediagxfb_scroll = USE_REDRAW;
		else if (! strcmp(this_opt, "memmove"))
			mediagxfb_scroll = USE_MEMMOVE;
		else if (!strncmp(this_opt, "font:", 5))
			strcpy(fb_info.fontname, this_opt+5);
	}
}

static int mediagxfb_switch(int con, struct fb_info *info)
{
	/* Do we have to save the colormap? */
	if (fb_display[currcon].cmap.len)
		fb_get_cmap(&fb_display[currcon].cmap, &fb_display[currcon].var, 1,
			    mediagx_getcolreg, info);
	
	currcon = con;
	/* Install new colormap */
	do_install_cmap(con, info);
	return 0;
}

/* 0 unblank, 1 blank, 2 no vsync, 3 no hsync, 4 off */

static void mediagxfb_blank(int blank, struct fb_info *info)
{
	/* Not supported */
}

__initfunc(void mediagxfb_init(void))
{
	int i,j;
	struct pci_dev *pdev=pci_devices, *sdev=NULL;
	int mediaGX=0;
	int support=0;
	unsigned long flags;
	
	u8 index, lock, gcr;
	u32 ramsize = 0;
	
	while(pdev != NULL)
	{
		if(pdev->vendor!=PCI_VENDOR_ID_CYRIX)
		{
			pdev=pdev->next;
			continue;
		}
		switch(pdev->device)
		{
			case PCI_DEVICE_ID_CYRIX_PCI_MASTER:
				mediaGX=1;
				break;
			case PCI_DEVICE_ID_CYRIX_5510:
				support=1;
				sdev=pdev;
				break;
			case PCI_DEVICE_ID_CYRIX_5520:
				support=2;
				sdev=pdev;
				break;
			case PCI_DEVICE_ID_CYRIX_5530_VIDEO:
				support=3;
				sdev=pdev;
				break;
		}
		pdev=pdev->next;
	}
	
	
	if(!mediaGX)
	{
		printk(KERN_ERR "MediaGX processor not found.\n");
		return;
	}
	
	if(support==0)
	{
		printk(KERN_ERR "MediaGX processor found, but no known video support chip.\n");
		return;
	}
	
	if(support!=2)
	{
		printk(KERN_ERR "Cyrix5520 support tested only.\n");
		return;
	}
	
	
	outb(GX_IOIDX_DIR0, GX_IOPORT_INDEX);
	outb(GX_IOIDX_GCR, GX_IOPORT_INDEX);
	
	gcr = inb(GX_IOPORT_DATA);
	
	/*
	 *	Perform the final MediaGX test ritual
	 */
	
	save_flags(flags);
	cli();
	 
	index = inb(0x3D4);
	outb(0x30, 0x3D4);
	lock = inb(0x3D5);
	
	outb(0, 0x3D5);
	
	if(inb(0x3D5) == 0xFF)
	{
		outb(0x57, 0x3D5);
		outb(0x4C, 0x3D5);
		
		if(inb(0x3D5)==0)
		{
			if(inb(0x3D5)==0)
			{
				outb(0x3E, 0x3D4);
				ramsize = (inb(0x3D5)-1)<<16;
			}
		}
	}
	
	outb(lock, 0x3D5);
	outb(index, 0x3D4);
	
	restore_flags(flags);
	
	if(ramsize==0)
	{
		printk(KERN_ERR "Cyrix: Failed to read SoftVGA configuration.\n");
		return;
	}
	
	/*
	 *	Ok all looks cool
	 */
	
	
	video_base          = (void *)((gcr&3)<<30);
	
	/*
	 *	Line length must be multiples of 1024
	 */
	video_linelength    = video_width/(video_bpp/8);
	video_linelength    = (video_linelength+0x3FF)&~0x3FF;
	video_size          = ramsize;
	vcomp_base	    = video_base+ramsize-1024;
	
	video_visual = (video_bpp == 8) ?
		FB_VISUAL_PSEUDOCOLOR : FB_VISUAL_DIRECTCOLOR;
        video_vbase = ioremap((unsigned long)video_base+0x00800000, video_size);
        
        
        if(video_vbase==NULL)
        {
        	printk(KERN_ERR "Video remap failed.\n");
        	return;
        }
        
        vcomp_vbase = video_vbase+ramsize;
        
        /*
         *	Map the support space
         */
         
	mediagx_base  = ioremap(video_base, 0x20000);
	
	if(mediagx_base==NULL)
	{
		printk(KERN_ERR "Unable to map video controller.\n");
		return;
	}
	
	support_chip = support;
	
	m5520_base = mediagx_base+0x10000;

	if(support == 2)
	{
		u8 cfg;
		/*
		 *	Map the 5520/30video
		 */
		m5520_pbase=sdev->base_address[3];
		pcibios_read_config_byte (sdev->bus->number, sdev->devfn, 0x42, &cfg);
		pcibios_write_config_byte (sdev->bus->number, sdev->devfn, 0x42, cfg | 9);
		printk("5520 support chip at 0x%X\n", m5520_pbase);
	}
	printk("mediagxfb_: %dx%dx%d, linelength=%d\n",
	       video_width, video_height, video_bpp, video_linelength);
	printk("mediagxfb_: framebuffer at 0x%p, mapped to 0x%p, size %dK\n",
	       video_base, video_vbase, video_size>>10);
	       
	if (mediagxfb_scroll == USE_REDRAW)  printk("mediagxfb_: scrolling=redraw\n");
	if (mediagxfb_scroll == USE_MEMMOVE) printk("mediagxfb_: scrolling=memmove\n");
	 
	mediagxfb_defined.xres=video_width;
	mediagxfb_defined.yres=video_height;
	mediagxfb_defined.xres_virtual=video_width;
	mediagxfb_defined.yres_virtual=video_height;
	mediagxfb_defined.bits_per_pixel=video_bpp;

	if (video_bpp > 8) {
		mediagxfb_defined.red.offset    = 12;
		mediagxfb_defined.red.length    = 6;
		mediagxfb_defined.green.offset  = 6;
		mediagxfb_defined.green.length  = 6;
		mediagxfb_defined.blue.offset   = 0;
		mediagxfb_defined.blue.length   = 6;
		mediagxfb_defined.transp.offset = 0;
		mediagxfb_defined.transp.length = 0;
		video_cmap_len = 16;
	} else {
		mediagxfb_defined.red.length   = 6;
		mediagxfb_defined.green.length = 6;
		mediagxfb_defined.blue.length  = 6;
		for(i = 0; i < 16; i++) {
			j = color_table[i];
			palette[i].red   = default_red[j];
			palette[i].green = default_grn[j];
			palette[i].blue  = default_blu[j];
		}
		video_cmap_len = 256;
	}
//	request_region(0x3c0, 32, "vga+");
	
	
	strcpy(fb_info.modename, "Cyrix MediaGX");
	fb_info.changevar = NULL;
	fb_info.node = -1;
	fb_info.fbops = &mediagxfb_ops;
	fb_info.disp=&disp;
	fb_info.switch_con=&mediagxfb_switch;
	fb_info.updatevar=&fb_update_var;
	fb_info.blank=&mediagxfb_blank;
	mediagxfb_get_var(&disp.var, -1, &fb_info);
	mediagxfb_set_disp(-1);

	if (register_framebuffer(&fb_info)<0)
		return;

//	printk("fb%d: %s frame buffer device\n",
//	       GET_FB_IDX(fb_info.node), fb_info.modename);
}


int init_module(void)
{
	mediagxfb_init();
	return 0;
}

void cleanup_module(void)
{
	modes[0].tcfg|=DC_TCFG_BLNK|DC_TCFG_BKRT;
	mediagx_video_mode(0);
        if(video_vbase)
        	iounmap(video_vbase);
        if(mediagx_base)
        	iounmap(mediagx_base);
        unregister_framebuffer(&fb_info);
}
